#pragma once
#include <vector>
#include <string>
#include <J3ML/J3ML.hpp>

class BaseEntity {
protected:
    ///If the entity has a parent entity, It's here.
    BaseEntity* parent;
    ///Entity list of child entities.
    std::vector<BaseEntity*> children;
public:
    u32 ticksAlive;
    bool draw = true;
    std::string uuid;
    std::string name;
    bool alive;
    virtual std::vector<std::string> GetEntityUUIDList() const { return std::vector<std::string> { "" }; }

    BaseEntity() = default;
    std::vector<BaseEntity*> GetChildren();
    ///Sets a given entity as this entities parent.
    virtual void SetParent(BaseEntity* parent);
    bool IsDescendantOf(BaseEntity* ancestor);
    bool IsAncestorOf(BaseEntity* descendant);
    std::vector<BaseEntity*> GetDescendants();
    std::vector<BaseEntity*> GetAncestors();
    [[nodiscard]] virtual BaseEntity* GetParent() const { return parent;}
    BaseEntity* FindFirstChild(std::string search_name);
    [[nodiscard]] BaseEntity* GetFamilyTreeRoot() const;
    BaseEntity* Add(BaseEntity* newChild);

    void DescendantAdded(BaseEntity* ent) {}
    void DescendantRemoved(BaseEntity* ent) {}
    void AncestorAdded(BaseEntity* ent) {}
    void AncestorRemoved(BaseEntity* ent) {}

    template <class T>
    void SerializeMemberData(T& archive){ archive & uuid.c_str() & name.c_str() & alive; }

    template <class T>
    void Serialize(T& archive) {
        SerializeMemberData(archive);
        archive & GetEntityUUIDList();
    }

    // TODO: Constrain to DerivedEntityType
    template <typename T>
    T* FindFirstChildOfType() const
    {
        for (auto& child : children) {
            T* p = dynamic_cast<T>(child);

            if (p != nullptr)
                return p;
        }
        return nullptr;
    }

    virtual void pre_render()  {}
    virtual void post_render() {}
    virtual void jglRenderPass() {}
    ///The default rendering routine. Works for 99% of cases.
    virtual void render() {}
    virtual void update(float elapsed) {}
    virtual void ticc(int tics) { ticksAlive++; }
};